# Linux Application Checker
# Application Analysis Module (Analysis.pm)
#
# Copyright (C) 2007-2008 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Analysis;

use strict;
use File::Basename;
use File::Temp qw/ tempfile tempdir /;

our @ISA = 'Exporter';
our @EXPORT = ( qw(
    ) );
#-----------------------------------------------------------------------

our @all_lsb = ("3.0", "3.1", "3.2", "4.0");

#-----------------------------------------------------------------------
# Global data structures

## Distributions
# $distrs->{$id} = {ID, NAME, VER};
our $all_distrs = {};

## Application Modules:
# $app_modules->{$module} = {
#       NAME => $module, FILE, ABITAG,
#       NEEDED_LIBS => [$soname],
#       NEEDED_LIBS_IMPL => [$soname], # implicitly needed libs, i.e. needed by needed libs.
#       NEEDED_INTS => {"$Iname\@$Iver" => $reqint_obj},
#       FOUND_LIBS->{$soname}{$Did} => $distrlib_obj,
#       FOUND_INTS->{$fullname}{$Did} => $reqlib,
#   };
our $app_modules = {}; 

## Required Libraries:
# $reqlibs->{$soname} -> {
#       SONAME,
#       LSBID, APPEAREDIN, MANDATORYSINCE, WITHDRAWNIN, DEPRECATEDSINCE, CANDIDATEFOR,
#       ATTR => {$LAtype => $LAvalue},
#       APPROVED,
#       ALL_FOUND_INTS => {$fullname => 1},
#   };
our $reqlibs = {};

## Libraries provided by distributions:
# $distrlibs->{$soname}{$distr}[] -> {
#       SONAME, ABITAG, 
#       POS, DISTR_ID, PATH,
#       IFACE_INTERVALS => [ [$first, $last] ]
#   };
my $distrlibs = {};

## Libraries scanned from a current distribution (to get some info on unknown libraries)
# $loclibs->{$soname}{$Iname} = 1;
our $loclibs = {};

## Required Interfaces:
# $reqints->{$Iname}{$ver} -> {
#       NAME, VER, UNMANGLED,
#       IDX_KEY, OTHER_VER_FOUND, RIIDs => [@RIIDs],
#       LSB_LIBRARY, APPEAREDIN, MANDATORYSINCE, WITHDRAWNIN, DEPRECATEDSINCE, CANDIDATEFOR,
#       ATTR => {$LAtype => $LAvalue},
#   };
our $reqints = {};

## Results of library analysis
# $libs->{$soname} -> { SONAME,
#		APP_MODULES => [$app_module],
#	    DISTRS => {$Did => 1},
#       INTS_USED,
#       LSBID, APPEAREDIN, MANDATORYSINCE, WITHDRAWNIN, DEPRECATEDSINCE, CANDIDATEFOR,
#       ATTR => {$LAtype => $LAvalue},
#   };
our $libs = {};
our $libs_impl = {};

# Library Knowledge Base
# $libKB->[$LKBid] = $text;
# $libKBDistr->{$soname}{$LKBid}{$group}{$dID} = 1;
our $libKB = [];
our $libKBDistr = {};

## Results of interface analysis
# $ints->{$Iname}{$ver} -> {
#       NAME, VER, UNMANGLED,
#       LIBRARIES => [],
#		APP_MODULES => [],
#	    OTHER_VER_FOUND, DISTRS => {$Did => $lib_with_path}
#       LSB_LIBRARY, APPEAREDIN, MANDATORYSINCE, WITHDRAWNIN, DEPRECATEDSINCE, CANDIDATEFOR,
#       ATTR => {$LAtype => $LAvalue},
#   };
our $ints = {};

# Interface Knowledge Base
# $intKB->[$IKBid] = $text;
# $intKBDistr->{$Iid}{$IKBid}{$group}{$dID} = 1;
my $intKB = [];
my $intKBDistr = {};


#-----------------------------------------------------------------------

my $crc32_poly = 0xEDB88320;

sub _crc32 {
	my $comp = shift;
	for (my $cnt = 0; $cnt < 8; $cnt++) { $comp = $comp & 1 ? $crc32_poly ^ ($comp >> 1) : $comp >> 1; }
	return $comp;
}

sub crc32 {
	my ($tcmp) = @_;
	my $crc = 0xFFFFFFFF;
	foreach (split(//,$tcmp)) { $crc = (($crc>>8) & 0x00FFFFFF) ^ _crc32(($crc ^ ord($_)) & 0xFF); }
	return $crc^0xFFFFFFFF;
}

sub load_distrs {
    $all_distrs = read_distrs(@_);
    
    return $all_distrs;
}

# This function reads Distributions file, which contains distro ID, name and other details.
sub read_distrs {
    my ($filename) = @_;
    
    my $distrs = {};
    
    open FILE, $filename
        or do { print STDERR "Failed to open file '$filename'."; return undef; };
    
    while ( my $line = <FILE> ) {
        chomp $line;
        next if $line =~ /^#/; # skip comments
        
        # Format is simple: fields are separated by TAB.
        my ($id, $distr_name, $ver, $year, $vendor) = split /\t/, $line;

        $distrs->{$id} = { ID => $id, NAME => $distr_name, VER => $ver, YEAR => $year, VENDOR => $vendor };
    }
    
    return $distrs;
}

#-----------------------------------------------------------------------

# Stores all the results into 3 files: libs info, library attrs, library knowledge base parts.
sub lib_dump {
    my ($libs_file, $attrs_file, $kb_file) = @_;
    
    open FILE, ">".$libs_file
        or die "Failed to open file '$libs_file'.";
    
    open ATTRS_FILE, ">".$attrs_file
        or die "Failed to open file '$attrs_file'.";

    my $lib_modules = {};
    foreach my $app_module ( keys %$app_modules ) {
        foreach my $soname ( @{$app_modules->{$app_module}{NEEDED_LIBS}} ) {
            $lib_modules->{$soname}{$app_module} = 1;
        }
    }
    
    foreach my $soname ( sort keys %$reqlibs ) {
        my $lib = $reqlibs->{$soname};
        
        my $approved  = ( $lib->{APPROVED} ? 1 : 0 );
        my $ints_used = ( scalar(keys %{$lib->{ALL_FOUND_INTS}}) or 0 );
        
        my %distrs = ();
        foreach my $Did ( keys %$all_distrs ) {
            foreach my $app_module ( values %$app_modules ) {
                next if !$app_module->{FOUND_LIBS}{$soname};
                my $val = $app_module->{FOUND_LIBS}{$soname}{$Did};
                if ( !$val ) {
                    delete $distrs{$Did} if exists $distrs{$Did};
                    last;
                }
                $distrs{$Did} = 1;
            }
        }
        my $distrs = join ",", keys %distrs;
        
        my $modules = "";
        if ( $lib_modules->{$soname} ) {
            $modules = join ",", sort keys %{$lib_modules->{$soname}};
        }
        
        my $lsb_info = "";
        if ( $lib->{LSBID} ) {
            
            $lsb_info = join "\t", map {defined and $_ or ""}
                    @$lib{qw/ APPEAREDIN MANDATORYSINCE WITHDRAWNIN DEPRECATEDSINCE CANDIDATEFOR /};
        }
        if ( $lib->{ATTR} ) {
        # ATTRS
            foreach my $LAtype ( keys %{$lib->{ATTR}} ) {
                my $LAvalue = $lib->{ATTR}{$LAtype};
                    
                $LAvalue =~ s/[\n\r]/\x81/g; # Use \x81 instead of \n
                
                print ATTRS_FILE join("\t", $soname, $LAtype, $LAvalue);
                print ATTRS_FILE "\n";
            }
        }        
        
        print FILE join( "\t", $soname, $approved, $ints_used, $modules, $distrs, $lsb_info );
        print FILE "\n";
    }
    
    close ATTRS_FILE;
    
    close FILE;

    # Lib Knowledge Base
    open FILE, ">".$kb_file
        or die "Failed to open file '$kb_file'.";
    
    foreach (my $LKBid=0; $LKBid < @$libKB; $LKBid++ ) {
		print FILE "\n";
		print FILE $libKB->[$LKBid]."\n";
		
		foreach my $soname ( keys %$libKBDistr ) {
		    next if !$libKBDistr->{$soname}{$LKBid};
			print FILE "LIB: $soname ";
			my @groups = ();
			foreach my $group ( keys %{$libKBDistr->{$soname}{$LKBid}} ) {
				push @groups, join ",", @{$libKBDistr->{$soname}{$LKBid}{$group}};
			}
			print FILE join ";", @groups;
			print FILE "\n";
		}
	}
    
	close FILE;
}

sub int_dump {
    my ($ints_file, $attrs_file, $kb_file) = @_;
    
    open FILE, ">".$ints_file
        or die "Failed to open file '$ints_file'.";
    
    open ATTRS_FILE, ">".$attrs_file
        or die "Failed to open file '$attrs_file'.";

    my $int_modules = {};
    foreach my $app_module ( keys %$app_modules ) {
        next if not $app_modules->{$app_module}{NEEDED_INTS};
        foreach my $int_obj ( values %{$app_modules->{$app_module}{NEEDED_INTS}} ) {
            arr_ref_push( $int_modules->{ $int_obj->{NAME} }{ $int_obj->{VER} }, $app_module );
        }
    }
    
    my $int_libraries = {}; # STRUCTURE: $int_libraries->{$fullname}{$soname} = 1;
    foreach my $libobj ( values %$reqlibs ) {
        next if !$libobj->{ALL_FOUND_INTS};
        foreach my $fullname ( keys %{$libobj->{ALL_FOUND_INTS}} ) {
            $int_libraries->{$fullname}{$libobj->{SONAME}} = 1;
        }
    }
    
    foreach my $Iname ( sort keys %$reqints ) {
        foreach my $ver ( sort keys %{$reqints->{$Iname}} ) {
            my $int_obj = $reqints->{$Iname}{$ver};
            
            my $unmangled = ( $int_obj->{UNMANGLED} or "" );
            
            my $other_ver = ( $int_obj->{OTHER_VER_FOUND} or "" );
            
            my $modules = "";
            if ( $int_modules->{$Iname}{$ver} ) {
                $modules = join ",", sort @{$int_modules->{$Iname}{$ver}};
            }
            
            my $distr_components = "";
            my %distr_components = ();
            my %distrs = ();
            foreach my $app_module ( values %$app_modules ) {
                my $fullname = $Iname.( $ver ? "@".$ver : '' );
                next if not $app_module->{FOUND_INTS}{$fullname};
                foreach my $Did ( keys %{$app_module->{FOUND_INTS}{$fullname}} ) {
                    my $reqlib = $app_module->{FOUND_INTS}{$fullname}{$Did};
                    next if !$reqlib;
                    
                    my $distrlib = $app_module->{FOUND_LIBS}{$reqlib->{SONAME}}{$Did};
                    next if !$distrlib;
                    
                    my $component = $distrlib->{PATH}."/".$distrlib->{SONAME};
                    $distr_components{$component}{$Did} = 1;
                    $distrs{$Did} = 1;
                }
            }
            
            my $fullname = $Iname.($ver ? '@'.$ver : '');
            my $libraries = "";
            if ( $int_libraries->{$fullname} ) {
                $libraries = join ",", keys %{$int_libraries->{$fullname}};
            } else {
                print STDERR "Warning: lib undefined for '$fullname'\n";
            }
            
            my @components = ();
            foreach my $component ( sort keys %distr_components ) {
                push @components, $component.":".(join ",", keys %{$distr_components{$component}});
            }
            $distr_components = join ";", @components;
            
            # Detect linkage contradictions
            # TODO: ^^^
            
            my $lsb_info = "";
            if ( $int_obj->{LSB_LIBRARY} ) {
                $lsb_info = join "\t", map {defined and $_ or ""}
                        @$int_obj{qw/ LSB_LIBRARY  APPEAREDIN MANDATORYSINCE WITHDRAWNIN
                                    DEPRECATEDSINCE CANDIDATEFOR /};
            }
            # ATTRS
            if ( $int_obj->{ATTR} ) {
                my $lsb_library = ( $int_obj->{LSB_LIBRARY} or "" );
                foreach my $LAtype ( keys %{$int_obj->{ATTR}} ) {
                    my $LAvalue = $int_obj->{ATTR}{$LAtype};
                    
                    $LAvalue =~ s/[\n\r]/\x81/g; # Use \x81 instead of \n
                    
                    print ATTRS_FILE join( "\t",
                            $Iname, $lsb_library, $LAtype, $LAvalue );
                    print ATTRS_FILE "\n";
                }
            }
            
            print FILE join( "\t", $Iname, $ver, $unmangled, $other_ver, $libraries, $modules,
                                $distr_components, $lsb_info );
            print FILE "\n";
        }
    }
    
    close ATTRS_FILE;
    
    close FILE;

    # TODO: Int Knowledge Base
}

# This functions loads results of library analysis
sub read_lib_dump {
    my ($libs_file) = @_;
    
    open FILE, $libs_file
        or die "Failed to open file '$libs_file'.";

    while( my $line = <FILE> ) {
        next if $line =~ /^#/;
        chomp $line;
        
        my ( $soname, $approved, $ints_used, $modules, $distrs, $lsb_info ) = split /\t/, $line, 6;
        
        my $lib_obj = {};
        $lib_obj->{SONAME} = $soname;
        $lib_obj->{APPROVED} = $approved;
        $lib_obj->{INTS_USED} = $ints_used;
        
        if ( $modules ) {
            $lib_obj->{APP_MODULES} = [split ",", $modules];
		}
		
        if ( $distrs ) {
            foreach my $dID ( split ",", $distrs ) {
                $lib_obj->{DISTRS}{$dID} = 1;
            }
		}
		if ( $lsb_info ) {
			@$lib_obj{qw/ APPEAREDIN MANDATORYSINCE WITHDRAWNIN DEPRECATEDSINCE CANDIDATEFOR /}
                    = split /\t/, $lsb_info;
		}
        
        if ( $modules ) {
            $libs->{$soname} = $lib_obj;
        } else {
            $libs_impl->{$soname} = $lib_obj; # implicitly requred library
        }
    }   
    close FILE;
}

# This functions loads results of interface analysis
sub read_int_dump {
    my ($ints_file) = @_;
    
    open FILE, $ints_file
        or die "Failed to open file '$ints_file'.";

    while( my $line = <FILE> ) {
        next if $line =~ /^#/;
        chomp $line;
        
        my ( $Iname, $ver, $unmangled, $other_ver, $libraries, $modules,
                    $distr_components, $lsb_info )
            = split /\t/, $line, 8;
        
        my $int_obj = {};
        $int_obj->{NAME} = $Iname;
        $int_obj->{VER} = $ver;
        $int_obj->{UNMANGLED} = $unmangled if $unmangled;
        $int_obj->{OTHER_VER_FOUND} = $other_ver if $other_ver;
        $int_obj->{LIBRARIES} = [split /,/, $libraries];
        
        my $unapproved = 0;
        foreach my $soname ( @{$int_obj->{LIBRARIES}} ) {
			if (   $libs->{$soname} && !$libs->{$soname}{APPROVED} 
				or $libs_impl->{$soname} && !$libs_impl->{$soname}{APPROVED} )
			{
			    $unapproved++;
            }
        }
        if ( $unapproved && $unapproved == scalar @{$int_obj->{LIBRARIES}} ) {
            $int_obj->{UNAPPROVED} = 1;
        }
        
        if ( $modules ) {
        	$int_obj->{APP_MODULES} = [split ",", $modules];
		}
        if ( $distr_components ) {
        	for my $component_ids ( split ";", $distr_components ) {
        		my ($lib_with_path, $ids) = split ":", $component_ids;
        		foreach my $Did ( split ",", $ids ) {
        		    if ( $int_obj->{DISTRS}{$Did} ) {
                        $int_obj->{DISTRS}{$Did} .= ','.$lib_with_path;
                    } else {
        		        $int_obj->{DISTRS}{$Did} = $lib_with_path;
                    }
                }
			}
		}
		if ( $lsb_info ) {
			@$int_obj{qw/ LSB_LIBRARY  APPEAREDIN MANDATORYSINCE WITHDRAWNIN
                                DEPRECATEDSINCE CANDIDATEFOR /}
                    = split /\t/, $lsb_info;
		}
		
        $ints->{$Iname}{$ver} = $int_obj;
    }   
    close FILE;
}

sub read_LibAttrs {
	my ($filename, my $libs) = @_;
    
    open FILE, $filename
        or die "Failed to open file '$filename'.";
    
    while( my $line = <FILE> ) {
        next if $line =~ /^#/;
        chomp $line;
        # FORMAT: LAsoname LAtype LAvalue
        
        my ($LAsoname, $LAtype, $LAvalue) = split /\t/, $line;
        
        my $lib_obj = $libs->{$LAsoname};
        next if !$lib_obj;
        
        $LAvalue =~ s/\x81/\n/g; # \x81 is used instead of \n.
        
        if ( defined $lib_obj->{ATTR}->{$LAtype} ) {
            $lib_obj->{ATTR}->{$LAtype} .= " ".$LAvalue;
        } else {
            $lib_obj->{ATTR}->{$LAtype} = $LAvalue;
        }
    }
    close FILE;
}

# This function reads a libKB file, which contains information
#  from the library Knowledge Base.
sub read_libKB {
    my ($filename) = @_;
    
    my $LKBid = 0;
    
    open FILE, $filename
        or die "Failed to open file '$filename'.";
    
    my $mode = "text";
    my $text = "";
    while ( my $line = <FILE> ) {
        chomp $line;
        next if $line =~ /^#/; # skip comments
        
        # Format: a multiline text followed by one or more "LIB: " lines.
        # "LIB: " line specifies a library and few behaviour groups of distributions.
        if ( $mode eq "text" ) {
            if ( $line =~ /^LIB: / ) {
                $mode = "lib";
            }
        }
        elsif ( $mode eq "lib" ) {
            if ( $line !~ /^LIB: / ) {
                $mode = "text";
                ++$LKBid;
                $text = "";
            }
        }
        
        if ( $mode eq "text" ) {
            next if ( $line eq "" && $text eq "" ); # Skip empty lines in the beginning of the message
            
            $text .= "\n" if $text ne "";
            $text .= $line;
        }
        elsif ( $mode eq "lib" ) {
            if ( !defined $libKB->[$LKBid] ) {
                $libKB->[$LKBid] = $text;
            }
            
            my ( $tmp, $soname, $distrs_grouped ) = split /\s+/, $line, 3;
            
            $distrs_grouped =~ s/\s+//g;
            
            # Distro groups are separated by semicolon ';'
            my @groups = split /;/, $distrs_grouped;
            foreach my $i ( 0 .. $#groups ) {
                # Distros in a group are separated by comma ','
                my @IDs = split /,/, $groups[$i];
                if ( @IDs ) {
                    $libKBDistr->{$soname}{$LKBid}{$i+1} = [ @IDs ];
                }
            }
        }
    }
    close FILE;
}

sub read_LSB_Libs {
    my ($filename) = @_;
    
    open FILE, $filename
        or die "Failed to open file '$filename'.";

    while( my $line = <FILE> ) {
        next if $line =~ /^#/;
        chomp $line;
        # FORMAT: Lid  Lname  ALrunname  ALappearedin  SMmandatorysince  ALwithdrawnin  Ldeprecatedsince  Lcandidatefor
        
        my ( $Lid, $Lname, $ALrunname, $ALappearedin, $SMmandatorysince,
                $ALwithdrawnin, $Ldeprecatedsince, $Lcandidatefor ) = split /\t/, $line;
        
        my $lib_obj = $reqlibs->{$ALrunname};
        next if !$lib_obj;
        
        $lib_obj->{LSBID} = $Lid;
        $lib_obj->{APPEAREDIN} = $ALappearedin if $ALappearedin;
        $lib_obj->{MANDATORYSINCE} = $SMmandatorysince if $SMmandatorysince;
        $lib_obj->{WITHDRAWNIN} = $ALwithdrawnin if $ALwithdrawnin;
        $lib_obj->{DEPRECATEDSINCE} = $Ldeprecatedsince if $Ldeprecatedsince;
        $lib_obj->{CANDIDATEFOR} = $Lcandidatefor if $Lcandidatefor;
    }
    close( FILE );
}

#=======================================================================

# Functions for working with Interfaces-related data

sub read_LibIndex {
    my ($filename) = @_;
    
    open FILE, $filename
            or die "Failed to open file '$filename'.";
    
    while ( my $line = <FILE> ) {
        next if $line =~ /^#/;
        chomp $line;
        
        my ($pos, $soname, $distr, $path, $abitag) = split /,/, $line;
        
        next if not defined $reqlibs->{$soname};
        
        $reqlibs->{$soname}{APPROVED} = 1;
        
        $abitag = "" if not defined $abitag;
        
        my $distrlib_obj = { POS => $pos, SONAME => $soname,  DISTR_ID => $distr,
                PATH => $path, ABITAG => $abitag,
            };
        
        arr_ref_push( $distrlibs->{$soname}{$distr}, $distrlib_obj );
    }
    
    close FILE;
}

sub read_UnapprovedLibs {
    my ($filename) = @_;
    
    open FILE, $filename
            or die "Failed to open file '$filename'.";
    
    while ( my $line = <FILE> ) {
        next if $line =~ /^#/;
        chomp $line;

        my ($soname, $distr, $path, $abitag) = split /,/, $line;
        
        next if !$reqlibs->{$soname};
        
        $abitag = "" if not defined $abitag;
        
        my $distrlib_obj;
        
        @$distrlib_obj{qw/SONAME DISTR_ID PATH ABITAG/}
                = ($soname, $distr, $path, $abitag);
        
        arr_ref_push( $distrlibs->{$soname}{$distr}, $distrlib_obj );
    }
    
    close FILE;
}

# Read LSB interfaces list
sub read_LSB_Ints {
    my ($filename) = @_;
    
    open FILE, $filename
        or die "Failed to open file '$filename'.";

    while( my $line = <FILE> ) {
        next if $line =~ /^#/;
        chomp $line;
        
        # First, split out two fields only (Iname is wanted)
        my ( $Iname, $Ver, $other ) = split /\t/, $line, 3;
        
        # Check whether this interface is required by the Application
        next if !$reqints->{$Iname};
        
        # Look if this interface is used by application
        my $int_obj = $reqints->{$Iname}{$Ver};
        if ( not $int_obj ) {
            $int_obj = $reqints->{$Iname}{""};
        }
        next if not $int_obj;
        
        # Split the rest
        my ( $generic, $Ilibrary, $AIappearedin, $SMmandatorysince,
				$AIwithdrawnin, $AIdeprecatedsince, $Icandidatefor ) = split /\t/, $other;
        
        $Ilibrary = "" if not defined $Ilibrary;
        
        if ( defined $int_obj->{LSB_LIBRARY} ) { # Already found
            # Entry with a specified architecture overrides entry with a Generic arch
            next if ( $generic && !$int_obj->{GENERIC} );
            
            if ( !$generic && $int_obj->{GENERIC} ) {
                delete $int_obj->{GENERIC};
                delete $int_obj->{APPEAREDIN};
                delete $int_obj->{WITHDRAWNIN};
                delete $int_obj->{DEPRECATEDSINCE};
            }

            # An interface can appear, be withdrawn and appear again! >:-F (e.g. clock_gettime, fstatfs64, inet_aton, ...)            
            # Merge info
            if ( $AIappearedin ) {
                $int_obj->{APPEAREDIN} .= ',' if $int_obj->{APPEAREDIN};
                $int_obj->{APPEAREDIN} .= $AIappearedin;
            }
            if ( $AIwithdrawnin ) {
                $int_obj->{WITHDRAWNIN} .= ',' if $int_obj->{WITHDRAWNIN};
                $int_obj->{WITHDRAWNIN} .= $AIwithdrawnin;
            }
            if ( $AIdeprecatedsince ) {
                $int_obj->{DEPRECATEDSINCE} .= ',' if $int_obj->{DEPRECATEDSINCE};
                $int_obj->{DEPRECATEDSINCE} .= $AIdeprecatedsince;
            }
        } else {
            $int_obj->{LSB_LIBRARY} = $Ilibrary;
            $int_obj->{GENERIC} = $generic if $generic;
            $int_obj->{APPEAREDIN} = $AIappearedin if $AIappearedin;
            $int_obj->{MANDATORYSINCE} = $SMmandatorysince if $SMmandatorysince;
            $int_obj->{WITHDRAWNIN} = $AIwithdrawnin if $AIwithdrawnin;
            $int_obj->{DEPRECATEDSINCE} = $AIdeprecatedsince if $AIdeprecatedsince;
            $int_obj->{CANDIDATEFOR} = $Icandidatefor if $Icandidatefor;
        }
    }
    close( FILE );
    
    return undef;
}

sub read_IntAttrs {
	my ($attrs_file, $ints) = @_;
    
    open FILE, $attrs_file
        or die "Failed to open file '$attrs_file'.";
    
    while( my $line = <FILE> ) {
        next if $line =~ /^#/;
        chomp $line;
        # FORMAT: IAiname IAlibrary IAtype IAvalue
        
        my ($IAiname, $IAlibrary, $IAtype, $IAvalue) = split /\t/, $line;
        
        next if !$ints->{$IAiname};
        
        $IAvalue =~ s/\x81/\n/g; # \x81 is used instead of \n.
        
        foreach my $int_obj ( values %{$ints->{$IAiname}} ) {
            next if (  $int_obj->{LSB_LIBRARY} && $IAlibrary
                    && $int_obj->{LSB_LIBRARY} ne $IAlibrary );
            
            if ( defined $int_obj->{ATTR}->{$IAtype} ) {
                $int_obj->{ATTR}->{$IAtype} .= " ".$IAvalue;
            } else {
                $int_obj->{ATTR}->{$IAtype} = $IAvalue;
            }
        }
    }
    close FILE;
}

# Load index for interfaces from "Community Tables"
sub read_IIndex {
    my ($filename) = @_;
    
    my @int_index = ();
    
    open FILE, $filename
        or die "Failed to open file '$filename'.";
    
    # Format: line number is an index beginning from 1000;
    # A number in each line is a position in Interfaces file.
    
    my $idx = 0;
    while ( my $line = <FILE> ) {
        chomp $line;
        
        $int_index[$idx] = $line;
        $idx++;
    }
    close FILE;
    
    return \@int_index;
}

# Searches for each interface required by application in "Community" data.
sub find_interfaces {
    my ($filename, $int_index) = @_;
    
    my $interfaces_fh;
    open $interfaces_fh, $filename
        or die "Failed to open file '$filename'.";
    
    my @int_arr = ();
    
    foreach my $Iname ( keys %$reqints ) {
        foreach my $int_obj ( values %{$reqints->{$Iname}} ) {
            next if $int_obj->{IDX_KEY};
            
            my $key = substr( crc32($Iname), 0, 4);
            $key -= 1000; # Array offset: crc sums starts from 1000...
            
            $int_obj->{IDX_KEY} = $key;
            
            push @int_arr, $int_obj;
        }
    }
    
    print "!find_interfaces: ".scalar(keys %$reqints)."\n";
    
    my $last_pos = 0;
    my $prev_pos = 0;
    my $prev_name = "";
    foreach my $int_obj ( 
            # Interfaces to be searched for are enumerated in the same order as in the file.
            # So we could read the file incrementally. This should enhance the performance.
            sort { $a->{IDX_KEY} <=> $b->{IDX_KEY} or $a->{NAME} cmp $b->{NAME} }
                @int_arr ) 
    {
        my $name = $int_obj->{NAME};
        my $ver = $int_obj->{VER};
        
        my $key = $int_obj->{IDX_KEY};
        if ( not defined $key ) {   # Absent in the index
            next;
        }
        
        my $pos = $int_index->[$key];
        my $max_pos = $int_index->[$key+1];

        if ( $name eq $prev_name ) {
            # If the name is the same as at the previous iteration (the version may differ)
            # that search from the same position.
            $pos = $prev_pos;
        }
        elsif ( $last_pos > $pos ) {
            # Search from the position where stopped at the previous iteration.
            $pos = $last_pos;
        }
        
        my ($found_pos, $other_ver_found, @IDs) = find_interface( $interfaces_fh, $int_index, $name, $ver, $pos, $max_pos );
        $last_pos = $found_pos if @IDs;
        
        if ( !@IDs ) {
            if ( $other_ver_found ) {
                $int_obj->{OTHER_VER_FOUND} = 1;
            }
        }
        else {
            $int_obj->{RIIDs} = [sort {$a <=> $b} @IDs];
        }
        
        $prev_name = $name;
        $prev_pos = $pos;
    }
    
    close $interfaces_fh;
}

# This function searches fro interface ID in the Interfaces file.
# Since the file is very large (over 30Mb), the index (IIndex file) is used.
# The Interfaces file consists of groups of interfaces with same crc sum of their names.
# Interfaces in a group sorted alphabetically by name.

# This function returns position of the last found element and an array
# of found interfaces (there could be several records for one interface name).
# The position of the last found element is used as a beginning position
# for the further search.
sub find_interface {
    my ($fh, $int_index, $name, $ver, $pos, $max_pos) = @_;
    
    die if !defined $name;
    
    seek $fh, $pos, 0;
    
    my @result = ();
    
    my $other_ver_found = 0;
    
    my $line_len;
    my $prev_name = "";
    while ( my $line = <$fh> ) {
        #print $line if $name eq 'wxPanelNameStr';
        $line_len = length($line);
        
        last if $max_pos && $pos >= $max_pos;
        
        chomp $line;
        
        if ( $line =~ /^([^,;]*)(,([^,;]+))?;(\d+)$/ ) {
            my $name1 = $1;
            my $ver1 = $3;
            my $Iid = $4;
            
            if ( $name1 eq "" ) {
            	$name1 = $prev_name;
			} else {
				next if $name1 lt $name; # continue searching
				
				last if $name1 gt $name; # no more
				
				$prev_name = $name1;
			}
            
            if ( $ver ) { # Should also check the version
                if ( defined $ver1 && $ver1 ne $ver ) {
                    $other_ver_found = 1;
                    next; # Wrong version, continue searching
                }
            }
            
            # Gotcha it if reached this line.
            push @result, $Iid;
        }
        else {
            die "Interfaces: wrong line '$line'";
        }
    }
    continue {
        $pos += $line_len;
    }
    
    return ($pos, $other_ver_found, @result); # Not found till EOF
}
#-----------------------------------------------------------------------

sub read_upload_data {
    my ($filename, $architecture) = @_;
    
    my %provided_libs = ();
    my %provided_ints = ();

    open FILE, $filename
        or die "Failed to open file '$filename'.";
    
    my $symtab = 0;    # indicates that we are processing 'symtab' section
    my $dynsym = 0;    # indicates that we are processing 'dynsym' section
    my $cmd = 0;       # indicates that we are processing output of a command
    my $lib = 0;       # indicates that we are processing output of a shared object
    my $dynsec = 0;    # indicates that we are processing general information about dynamic section
    my $elfheader = 0; # indicates that we are processing general information about elf header
    my $skip = 0;      # indicates that we are skipping lines until 'Library' or 'Command' record

    my $elf_arch_bit = 0;  # do we deal with 32 or 64 bit ELF file

    my $cur_app_module; # Application module being processed
    
    while( <FILE> ) {
        chomp;
        
        if( ($dynsec or $dynsym) and ( /^\S/ ) ) {
            $dynsec = $dynsym = 0;
        }

        if( /^ELF Header/ ) {
            $elfheader = 1;
            next;
        }
        
        if( $elfheader ) {
            if ( /^\s+Class\:\s*ELF(\d+)/ ) {
                $skip = 0;
                $elf_arch_bit = $1;
                next;
            }
            elsif ( /^\s+Machine\:\s*(.+)/ ) {
                $skip = 0;
                my $elf_arch = $1;

                if( !is_elf_arch_correct($architecture, $elf_arch, $elf_arch_bit) ) {
                    $skip = 1;
                }
                next;
            }
            elsif ( /^\S/ ) {
                $elfheader = 0;
                $elf_arch_bit = 0;
            }
        }

        if( /^\!Component/ ) {
            # new component, reset flags
            $symtab = $dynsym = $dynsec = $cmd = $skip = $lib = 0;
        }
        elsif( /^\!Library/ ) {
            # new shared object, reset flags
            $symtab = $dynsym = $dynsec = $cmd = $skip = 0;
            $lib=1;

            # The string should contain name, version, runname, path, ABI tag (optional) and so-name (optional) of the library
            if( /^\!Library '([^']+)' '([^']*)' '([^']+)' '([^']+)' '([^']*)' '([^']*)'/ ) {
                my $runname = $3;
                my $soname = $6;
                my $origin_name = "$4/$runname";
                
                my $app_abitag = $5;
                
                # Remember this lib as provided by the app
                $provided_libs{$runname} = 1;
                if( $runname =~ /(.+\.so)\./ ) {
                    $provided_libs{$1} = 1;
                }
                if( $soname ) {
                    $provided_libs{$soname} = 1;
                }
                
                $cur_app_module = $origin_name;
                
                my $real_file = $cur_app_module;
                if ( $cur_app_module =~ /([^\|]*)\|(.*)/ ) {
                    $real_file = $1;
                    $cur_app_module = $2;
                }
                
                $app_modules->{$cur_app_module} = { NAME => $cur_app_module, FILE => $real_file,
                        NEEDED_LIBS => [], NEEDED_LIBS_IMPL => [], NEEDED_INTS => {} };
                $app_modules->{$cur_app_module}{ABITAG} = $app_abitag if $app_abitag;
            }
            else {
                print STDERR "Incorrect library record at line $.:\n\t'$_'\n";
            }
        }
        elsif( /^Dynamic\ section/ ) {
            next if $skip;
            $symtab = $dynsym = 0;
            $dynsec=1;
        }
        elsif( /^\S/ && /'.dynsym'/ ) {
            next if $skip;
            $symtab = $dynsec = 0;
            $dynsym=1;
        }
        elsif( /^\S/ && /'.symtab'/ ) {
            next if $skip;
            $dynsym = $dynsec = 0;
            $symtab=1;
        }
        elsif( $dynsec && /NEEDED\)\s+Shared\s+library\:\s+\[((\w|-|\+|.)+)\]/ ) {
            next if $skip;

            my $fullname = $1;
            my $libname;

            if ( $fullname =~ m|/([^/]+)/?$| ) {
                $libname = $1;
            } else {
                $libname = $fullname;
            }
            
            my $lib_obj = $reqlibs->{$libname};
            if ( !$lib_obj ) {
                $lib_obj = { SONAME => $libname };
                $reqlibs->{$libname} = $lib_obj;
            }
            
            push @{$app_modules->{$cur_app_module}{NEEDED_LIBS}}, $libname;
        }
        elsif(  ($dynsym or $symtab) 
                && /\s*\d+:\s+\w*\s+\w+\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s((\w|@|\.)+)/ ) 
        {
            # (the line of 'readelf' output corresponding to the interface)
            next if $skip;

            # we also need to know if the interface is in the shared object
            # or in the library
            if( ($lib==0) and ($cmd==0) ) {
                print STDERR "Cannot determine if interface $1 is in executable or in shared object!\n";
                exit(1);
            }

            # do nothing with symtab (but there are some plans for the future)
            if( ($symtab == 1) or ($dynsym == 0) ) {
                next;
            }

            my $type = $1;     # FUNC, OBJECT etc.
            my $bind = $2;     # GLOBAL, WEAK or LOCAL
            my $vis = $3;      # Visibility
            my $Ndx = $4;	   # 
            my $fullname = $5; # name, maybe with version

            # Filter interfaces by their binding, type and visibility
            if( ($bind ne 'WEAK') and ($bind ne 'GLOBAL') ) {
                next;
            }
            if( ($type ne 'FUNC') and ($type ne 'OBJECT') and ($type ne 'COMMON') ) {
                next;
            }
            if( $vis ne 'DEFAULT' ) {
                next;
            }

            $fullname =~ s/\@\@/\@/;
            my ($Iname, $Iver) = split /\@/, $fullname;
            $Iver = "" if !$Iver;
            
            if( $Ndx eq "UND" ) {
                # Required interface
                
                my $reqint_obj = $reqints->{$Iname}{$Iver};
                if ( not $reqint_obj ) {
                    $reqint_obj = { NAME => $Iname, VER => $Iver };
                    $reqints->{$Iname}{$Iver} = $reqint_obj;
                }
                
                $app_modules->{$cur_app_module}{NEEDED_INTS}->{$fullname} = $reqint_obj;
            }
            else {
                # Internal interface
                $provided_ints{$fullname} = 1;
                $provided_ints{$Iname} = 2 if !$provided_ints{$Iname};
            }
        }
        elsif( /^\!Command/ ) {
            # The string should contain name and path
            if( /^\!Command '([^']+)' '([^']*)'/ ) {
                $cur_app_module = "$2";
                
                my $real_file = $cur_app_module;
                if ( $cur_app_module =~ /([^\|]*)\|(.*)/ ) {
                    $real_file = $1;
                    $cur_app_module = $2;
                }
                
                $app_modules->{$cur_app_module} = { NAME => $cur_app_module, FILE => $real_file,
                    NEEDED_LIBS => [], NEEDED_LIBS_IMPL => [], NEEDED_INTS => {} };
            }
            else {
                print STDERR "Incorrect command record at line $.:\n\t'$_'\n";
            }

            # new command, reset flags
            $symtab = $dynsym = $dynsec = $lib = $skip = 0;
            $cmd=1;
        }
        #elsif ( /^\!Script/ ) {
        #	# Do nothing
        #}
    }
    
    my $tempdir = tempdir( 'appchk-analysis-XXXXXX', CLEANUP => 1 );
    
    # get implicit dependencies
    foreach my $app_module (values %$app_modules ) {    
        $app_module->{NEEDED_LIBS_IMPL} = [];
        
        my %needed_libs;
        $needed_libs{ $_ } = 1 foreach ( @{$app_module->{NEEDED_LIBS}} );
        
        my $filename = $app_module->{FILE};
        
        if ( !-x $filename  ) {
            # Copy the file to obtain the X permission to be able to use ldd with this file.
            if ( system( "cp '$filename' '$tempdir'" ) == 0 ) {
                $filename = $tempdir."/".basename($filename);
            }
        }
        open PIPE, "ldd '$filename' |";
        
        while ( my $line = <PIPE> ) {
		    chomp $line;
		    
		    if ( $line =~ /\s+(\S+)\s/ ) {
		        my $soname = $1;
		        next if defined $needed_libs{$soname}; # already in NEEDED
		        
		        my $lib_obj = $reqlibs->{$soname};
		        if ( !$lib_obj ) {
		            $lib_obj = {SONAME => $soname};
		            $reqlibs->{$soname} = $lib_obj;
                }
		        # add to implicit NEEDED
		        push @{$app_module->{NEEDED_LIBS_IMPL}}, $soname;
            }
        }
        
        close PIPE;
    }
    
    $tempdir and system("rm -r '$tempdir'");
    
    # Delete provided libs and shared ints
    foreach my $app_module ( values %$app_modules ) {
        # Delete provided libs
        @{$app_module->{NEEDED_LIBS}} = grep {!$provided_libs{$_}} @{$app_module->{NEEDED_LIBS}};
        @{$app_module->{NEEDED_LIBS_IMPL}} = grep {!$provided_libs{$_}} @{$app_module->{NEEDED_LIBS_IMPL}};
        
        # Delete shared ints
        if ( defined $app_module->{NEEDED_INTS} ) {
            foreach my $fullname ( keys %{$app_module->{NEEDED_INTS}} ) {
                my ($Iname) = split /@/, $fullname;
                if (   $provided_ints{$fullname} # fullname matches
                    || $provided_ints{$Iname} && $provided_ints{$Iname} == 1 ) # interface without version is provided
                {
                    delete $app_module->{NEEDED_INTS}->{$fullname};
                }
            }
        }
    }
    foreach my $soname ( keys %$reqlibs ) {
        if ( $provided_libs{$soname} ) {
            delete $reqlibs->{$soname};
        }
    }
    
    foreach my $Iname ( keys %$reqints ) {
        foreach my $ver ( keys %{$reqints->{$Iname}} ) {
            my $fullname = $Iname.($ver ? '@'.$ver : '');
            if (   $provided_ints{$fullname} # interface with proper version is provided
                || $provided_ints{$Iname} && $provided_ints{$Iname} == 1 )  # interface without version is provided
            {
                delete $reqints->{$Iname}{$ver};
                if ( !%{$reqints->{$Iname}} ) {
                    delete $reqints->{$Iname};
                }
            }
        }
    }
    
    close FILE;
}

sub find_libifaces {
    my ($filename) = @_;

    my %by_pos = ();
    foreach my $soname ( keys %$distrlibs ) {
        foreach my $Did ( keys %{$distrlibs->{$soname}} ) {
            foreach my $distrlib ( @{$distrlibs->{$soname}{$Did}} ) {
                next if not defined $distrlib->{POS};
                $by_pos{ $distrlib->{POS} } = $distrlib;
            }
        }
    }
    
    my $libifaces_fh;
    open $libifaces_fh, $filename
        or die "Failed to open file '$filename'.";
    
    foreach my $pos ( sort keys %by_pos ) {
        read_LibIfaces( $libifaces_fh, $pos, $by_pos{$pos} );
    }

    close $libifaces_fh;
}

sub read_LibIfaces {
    my ($fh, $pos, $distrlib_obj) = @_;
    
    seek $fh, $pos, 0
        or die "read_LibIfaces: seek to '$pos' failed";
    
    my $pairs = []; # $pairs = [ [$first, $last] ];
    
    # Read Interface IDs
    my $line = <$fh>;
	chomp $line;
	last if $line eq ""; # next library
	
	# Interface ID intervals
	my $first;
	my $last;
	if ( $line =~ s/^\d+// ) {
		$first = $last = int($&);
	} else {
		die "No base in line $line"; # DBG:
	}
	
	while ( $line =~ s/^([\+\-])(\d*)// ) {
		my $pm = $1;
		my $d = $2; 
		$d = 1 if !$d;
		
		if ( $pm eq '+' ) {
			$last = $first + $d - 1;
			arr_ref_push( $pairs, [$first, $last] );
		}
		elsif ( $pm eq '-' ) {
			$last = $first = $last + $d + 1;
		}
		else {
			die "Syntax error"; # should not happen
		}
	}
    
    $distrlib_obj->{IFACE_INTERVALS} = $pairs;
    
    return undef;
}

sub link_libs {
    my ($reqlibs) = @_;
    
    #print "link_libs: ".scalar(keys %$reqlibs)."\n"; # DBG:
    
    my $cache = {};
    
    foreach my $app_module ( values %$app_modules ) {
        my $abitag = ($app_module->{ABITAG} or "");
        
        foreach my $soname ( @{$app_module->{NEEDED_LIBS}}, @{$app_module->{NEEDED_LIBS_IMPL}} ) {
            
            my $distrlibs_d = $distrlibs->{$soname};
            next if !$distrlibs_d;
            
            my $cache_d = undef;
            if ( defined $cache->{$abitag} && defined $cache->{$abitag}{$soname} ) {
                $cache_d = $cache->{$abitag}{$soname};
            }
            foreach my $Did ( keys %$distrlibs_d ) {
                if ( $cache_d && $cache_d->{$Did} ) {
                    $app_module->{FOUND_LIBS}{$soname}{$Did} = $cache_d->{$Did};
                    next;
                }
                
                my $libs = $distrlibs_d->{$Did};
                
                next if !$libs || !@$libs;
                
                my $lib = linker_heuristic($libs, $abitag);
                
                if ( $lib ) {
                    $app_module->{FOUND_LIBS}{$soname}{$Did} = $lib;
                }
                
                $cache->{$abitag}{$soname}{$Did} = $lib;
            }
        }
    }
}

sub parse_abitag( $ ) {
    my ($abitag) = @_;
    
    if ( !defined $abitag || $abitag eq "" ) {
        return (0, 0, 0);
    }
    
    if ( $abitag =~ /^Linux\s+(\d+)\.(\d+)\.(\d+)\s*$/ ) {
        return ($1, $2, $3);
    }
    else {
        print STDERR "Warning! Can't parse ABI-tag '$abitag'\n";
        return (0, 0, 0);
    }
}

sub abitag_ge ( $ $ ) {
    my ($a, $b) = @_;
    
    if (   $a->[0] >  $b->[0]
        || $a->[0] == $b->[0] && $a->[1] >  $b->[1]
        || $a->[0] == $b->[0] && $a->[1] == $b->[1] && $a->[2] >= $b->[2] ) 
    {
        return 1;
    }
    
    return 0;
}

sub linker_heuristic {
    my ($libs, $app_abitag) = @_;
    
    if ( scalar @$libs == 0 ) { return ''; }
    if ( scalar @$libs == 1 ) { return $libs->[0]; }
    
    my @libs = @$libs; # Make a working copy
    my @libs_copy; # Temporary var
    
    # ( Explanation on ABI tags by Ulrich Drepper:
    #   http://people.redhat.com/drepper/assumekernel.html )
    
    if ( $app_abitag ) {        
        my @app_abitag = parse_abitag($app_abitag);
        
        @libs_copy = ();
        
        my $select_min = 1;
        
        # Filter by abitag
        foreach my $lib ( @libs ) {
            if ( $lib->{ABITAG} ) {
                my @lib_abitag = parse_abitag( $lib->{ABITAG} );
            
                next if not abitag_ge( \@lib_abitag, \@app_abitag );
            } else {
                $select_min = 0;
            }
            
            push @libs_copy, $lib;
        }
        
        if ( scalar @libs_copy == 1 ) { return $libs->[0]; }
        
        if ( @libs_copy ) {
            @libs = @libs_copy;
        } else {
            $select_min = 0;
        }
        
        if ( $select_min ) {
            # Select min of filtered abitags
            my @min_abitag = (99,99,99);
            my $min_abitag = "";
            foreach my $lib ( @libs ) {
                my @lib_abitag = parse_abitag( $lib->{ABITAG} );
                
                if ( not abitag_ge( \@lib_abitag, \@min_abitag )  ) {                
                    @min_abitag = @lib_abitag;
                    $min_abitag = $lib->{ABITAG};
                }
            }
            @libs = grep {$_->{ABITAG} eq $min_abitag} @libs;
            
            if ( scalar @libs == 1 ) { return $libs[0]; }
        }
    }

    # Linker uses /usr/lib and /lib after all others, so try to select something except them
    @libs_copy = grep {$_->{PATH} !~ '^/lib(/|32/|64/|$)' && $_->{PATH} !~ '^/usr/lib(/|32/|64/|$)'} @libs;
    if ( @libs_copy ) {
        @libs = @libs_copy; # select from this new subset
    }
    else { # There were only /usr/lib and /lib, use /lib.
        @libs_copy = grep { $_->{PATH} =~ '^/lib(/|32/|64/|$)' } @libs;
        if ( @libs_copy ) {
            @libs = @libs_copy;
        }
    }
    
    if ( scalar @libs == 1 ) { return $libs[0]; }
    
    # Try /lib/tls priority
    @libs_copy = grep {$_->{PATH} =~ m{.*/tls(/|$)} } @libs;
    if ( @libs_copy ) {
        @libs = @libs_copy;
    }
    
    if ( scalar @libs == 1 ) { return $libs[0]; }
    
    # Try /lib/i686 priority
    @libs_copy = grep {$_->{PATH} =~ m{.*/i[3-6]86(/|$)} } @libs;
    if ( @libs_copy ) {
        @libs = @libs_copy;
    }
    
    if ( scalar @libs == 1 ) { return $libs[0]; }
    
    # Pick any
    @libs = reverse sort {$a->{PATH} cmp $b->{PATH}} @libs;
    return $libs[0];
}

sub link_ints {
    foreach my $app_module (values %$app_modules ) {
        foreach my $fullname ( keys %{$app_module->{NEEDED_INTS}} ) {
            my $riids = $app_module->{NEEDED_INTS}{$fullname}->{RIIDs};
            # Even if $riids not found
            foreach my $Did ( keys %$all_distrs ) {
                
                local *search = sub {
                    my ($needed_libs) = @_;
                    
                    my $found_libs = $app_module->{FOUND_LIBS};
                    
                    foreach my $soname ( @$needed_libs ) {
                        my $distrlib = $found_libs->{$soname}{$Did};
                        
                        if ( $riids && $distrlib && $distrlib->{IFACE_INTERVALS} ) {
                            # Search in the $distrlib
                            my $pairs = $distrlib->{IFACE_INTERVALS};
                            
                            my $riidx = 0;
                            my $riid = $riids->[$riidx];
                            foreach my $pair ( @$pairs ) {
                                if ( $riid < $pair->[0] ) {
                                    $riidx++;
                                    last if $riidx > $#$riids;
                                    $riid = $riids->[$riidx];
                                    redo;
                                }
                                elsif ( $riid <= $pair->[1] ) {
                                    # gotcha!
                                    return ($soname, 1);
                                }
                            }
                        }
                        else { # Unknown library or iface
                        	# die "No IFACE_INTERVALS '$soname'" if $riids && $reqlibs->{$soname}{APPROVED}; # DBG:
                            # look in local lib
                            if ( $loclibs->{$soname}{$fullname} ) {
                                return ($soname, undef);
                            }
                        }
                    } # end of  foreach $soname ( ... )
                    return undef;
                }; # end of  local sub search()
                
                # search in NEEDED_LIBS
                my ($soname, $i) = search($app_module->{NEEDED_LIBS});
                if ( !$soname ) { # then search in NEEDED_LIBS_IMPL
                    ($soname, $i) = search($app_module->{NEEDED_LIBS_IMPL});
                }
                if ( $soname ) {
                    my $reqlib = $reqlibs->{$soname};
                    $reqlib->{ALL_FOUND_INTS}->{$fullname} = 1;
                    if ( $i ) {
                        $app_module->{FOUND_INTS}{$fullname}{$Did} = $reqlib;
                    }
                    next;
                }
            }
        }
    }
}

# Unmangle interface names
sub unmangle {
    # We make a list of mangled interface names,
    # then put it into a temporary file, then feed this file to c++filt
    # and read the unmangled name from its output.
    
    # Remember names order
    my $unmangle_order = [keys %$reqints];
    
    # Create a temporary file
    my ($tempfh, $tempfilename) = tempfile(); # from File::Temp
    foreach ( @$unmangle_order ) {
        print $tempfh "$_\n";
    }
    close $tempfh;
    
    # Call c++filt
    open CPPFILT, "cat '$tempfilename' | c++filt |";
    
    # Read the result from stdout
    foreach my $Iname ( @$unmangle_order ) {
        my $unmangled_name = <CPPFILT>;
        if ( !$unmangled_name ) {
            die "c++filt stopped when unmangled name of '$Iname' was expected." ;
        }
        chomp $unmangled_name;
        if ( $unmangled_name ne $Iname ) {
            foreach my $int_obj ( values %{$reqints->{$Iname}} ) {
                $int_obj->{UNMANGLED} = $unmangled_name;
            }
        }
    }
    close CPPFILT;
}

sub get_ldconfig {
    my ($arch) = @_;
    
	my $grep = "";
	my $neg = 0;
	
	if( $arch =~ /x86-64/i ) {
		$grep = "x86-64";
	}
	elsif( $arch =~ /IA32/i or $arch =~ /x86/i ) {
		# x86 apps will run on x86-64, but they will use 32bit libs
		$grep = "x86-64";
		$neg = 1;
	}
	elsif( $arch =~ /IA64/i ) {
		$grep = "IA-64";
	}
	elsif( $arch =~ /S390X/i ) {
		$grep = "64bit";
	}
	elsif( $arch =~ /PPC64/i ) {
		$grep = "64bit";
	}
	elsif( $arch =~ /S390/i ) {
		$grep = "64bit";
		$neg = 1;
	}

	open PIPE, "/sbin/ldconfig -p |";
	
	my $ld_config = {};
	
	while ( <PIPE> ) {
		chomp;
		
		my $test = /\b$grep\b/;
		my $wrong = 0;
		if ( $neg && $test or !$neg && !$test) {
			# Wrong bit number
			$wrong = 1;
		}
		
		if ( /\s+(\S+)\s[^=]*=>\s*(\S+)/ ) {
			my $libname = $1;
			my $realname = $2;
			
			if ( !$wrong || !$ld_config->{$libname} ) {
				$ld_config->{$libname} = $realname;
				# Also save short soname. It's better to find a wrong version of lib than nothing.
				$libname =~ s/(?<=\.so).*//;
				if ( !$ld_config->{$libname} ) {
					$ld_config->{$libname} = $realname;
				}
			}
		}
	}
	return $ld_config;
}

sub get_unknown_libs {
    my ($architecture) = @_;
    
    my $unknown_libs = {};
    
    foreach my $app_module (values %$app_modules ) {
        foreach my $soname ( @{$app_module->{NEEDED_LIBS}} ) {
            # Scan even known libs!
            #next if $distrlibs->{$soname} && $reqlibs->{$soname}{APPROVED};
            next if $loclibs->{$soname};
            
            $unknown_libs->{$soname} = 1;
        }
    }
    
    return if !%$unknown_libs;
    
    my $ld_config = get_ldconfig($architecture);
    
    foreach my $soname ( keys %$unknown_libs ) {
        scan_local_lib($soname, $ld_config);
    }
}

# Extract interfaces from a local library
sub scan_local_lib {
	my ($libname, $ld_config) = @_;
	
	my $realname = $ld_config->{$libname};
	if ( !$realname ) {
		# Try soname without version even if it's not an exact match
		$libname =~ s/(?<=\.so).*//;
		$realname = $ld_config->{$libname};
	}
	return if !$realname;
	
	# -WhlSsdA
	open( LIB_DATA, "LC_ALL=C readelf -Ws $realname | " );

	my $dynsym = 0;
	
	while( <LIB_DATA> ) {
		# We actually need only '.dynsym' section at the moment
		if( /^Symbol\s*table\s*'\.dynsym'/i ) {
			$dynsym = 1;
			next;
		}

		if( !$dynsym ) {
			next;
		}
		
		# Non-space in the beginning of the line; '.dynsym' section is finished
		if( /^\S/ ) {
			$dynsym = 0;
			last;
		}
		
		chomp;

		if( /\s*\d+:\s+\w*\s+\w+\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s((\w|@|\.)+)/ ) {
			# (the line of 'readelf' output corresponding to the interface)
			my $type=$1;     # FUNC, OBJECT etc.
			my $bind=$2;     # GLOBAL, WEAK or LOCAL
			my $vis=$3;      # Visibility
			my $Ndx=$4;		 #
			my $fullname=$5; # name, maybe with version
			
			# Filter interfaces by their binding, type and visibility
			if( ($bind ne 'WEAK') and ($bind ne 'GLOBAL') ) {
				next;
			}
			if( ($type ne 'FUNC') and ($type ne 'OBJECT') and ($type ne 'COMMON') ) {
				next;
			}
			if( ($vis ne 'DEFAULT') or ($Ndx eq "UND") ) {
				next;
			}

            $fullname =~ s/\@\@/\@/;
            
			$loclibs->{$libname}{$fullname} = 1;
			my ($shortname,$version) = split( "@", $fullname );
			$loclibs->{$libname}{$shortname} = 1;
		}
	}
	close( LIB_DATA );
}


sub arr_ref_push {
    if ( $_[0] ) {
        push @{$_[0]}, @_[1 .. $#_];
    } else {
        $_[0] = [ @_[1 .. $#_] ];
    }
}

sub is_elf_arch_correct($$$ ) {
	my($arch, $elf_arch, $elf_arch_bit) = ($_[0],$_[1],$_[2]);

	if( $arch =~ /x86-64/i ) {
		if( $elf_arch !~ /86-64/ ) {
			return 0;
		}
	}
	elsif( $arch =~ /IA64/i ) {
		if( $elf_arch !~ /IA-64/ ) {
			return 0;
		}
	}
	elsif( $arch =~ /S390X/i ) {
		if( ($elf_arch !~ /390/) or ($elf_arch_bit ne "64") ) {
			return 0;
		}
	}
	elsif( $arch =~ /PPC64/i ) {
		if( $elf_arch !~ /PowerPC 64-bit/ ) {
			return 0;
		}
	}
	elsif( $arch =~ /PPC32/i ) {
		if( ($elf_arch !~ /PowerPC/) or ($elf_arch =~ /PowerPC 64-bit/) or ($elf_arch =~ /PowerPC64/) ) {
			return 0;
		}
	}
	elsif( $arch =~ /x86/i ) {
		if( $elf_arch !~ /Intel 80386/ ) {
			return 0;
		}
	}
	elsif( $arch =~ /S390/i ) {
		if( ($elf_arch !~ /390/) or ($elf_arch_bit ne "32") ) {
			return 0;
		}
	}
	else {
	    print STDERR "is_elf_arch_correct(): Unknown arch: $arch\n";
    }

	return 1;
}
#-----------------------------------------------------------------------

sub distr_status {
	my ($Did) = @_;
	
	my $lib_statistics = {};
	my $int_statistics = {};
	
	foreach my $lib ( values %$libs ) {
		if ( $lib->{DISTRS}{$Did} ) { # Present
            $lib_statistics->{OK}++;
		}
		else { # Missed
		    if ( ! $lib->{APPROVED} ) {
				# Unknown lib
				$lib_statistics->{UNKNOWN}++;
			} 	
			else {
			    if ( ! $lib->{INTS_USED} ) {
                    # Zero ifaces used; known or unknown: UNNEEDED
                    $lib_statistics->{UNNEEDED}++;
                }
                else {
                    # Known missed lib
                    $lib_statistics->{MISSED}++;
                }
            }
		}
	}
	
	foreach my $int_vers ( values %$ints ) {
		foreach my $int ( values %$int_vers ) {
			if ( $int->{DISTRS}{$Did} ) { # Present
				$int_statistics->{OK}++;
			}
			else { # Missed
				if ( $int->{UNAPPROVED} ) {
					# Unknown int
					$int_statistics->{UNKNOWN}++;
				} 		
				else {
					# Known missed int
					$int_statistics->{MISSED}++;
				}
			}
		}
	}
	
	# Deduce the status:
	my $status;
	
	if ( $lib_statistics->{MISSED} || $int_statistics->{MISSED} ) {
		$status = 'MISSED';
	}
	elsif ( $lib_statistics->{UNKNOWN} || $int_statistics->{UNKNOWN} ) {
		$status = 'UNKNOWN';
	}
	elsif ( $lib_statistics->{UNNEEDED} ) {
		$status = 'UNNEEDED';
	}
	else {
		$status = 'OK';
	}
	
	return ($status, $lib_statistics, $int_statistics);
}

#-----------------------------------------------------------------------
1;
